#include "config.h"
#include "hardware.h"
#include "commands.h"
#include "i2c.h"

#define SERIAL_BAUD_RATE 115200

#define SERIALCOM Serial

#define COMMS_MONITOR_TIME_MS 5000 //I2C communications monitor grace/trigger time

// Defines for fast ADC reads
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

class AvrHardware : public Hardware {
private:
    // i2c address for node
    // Node 1 = 8, Node 2 = 10, Node 3 = 12, Node 4 = 14
    // Node 5 = 16, Node 6 = 18, Node 7 = 20, Node 8 = 22
    uint8_t i2cAddress = 6 + (NODE_NUMBER * 2);

#if defined(USE_I2C) && (!defined(NODE_NUMBER) || !NODE_NUMBER)
    // Configure the I2C address based on input-pin level.
    void configI2cAddress()
    {
        // current hardware selection
        pinMode(HARDWARE_SELECT_PIN_1, INPUT_PULLUP);
        pinMode(HARDWARE_SELECT_PIN_2, INPUT_PULLUP);
        pinMode(HARDWARE_SELECT_PIN_3, INPUT_PULLUP);
        // legacy selection - DEPRECATED
        pinMode(LEGACY_HARDWARE_SELECT_PIN_1, INPUT_PULLUP);
        pinMode(LEGACY_HARDWARE_SELECT_PIN_2, INPUT_PULLUP);
        pinMode(LEGACY_HARDWARE_SELECT_PIN_3, INPUT_PULLUP);
        pinMode(LEGACY_HARDWARE_SELECT_PIN_4, INPUT_PULLUP);
        pinMode(LEGACY_HARDWARE_SELECT_PIN_5, INPUT_PULLUP);

        delay(100);  // delay a bit a let pin levels settle before reading inputs

        // check if legacy spec pins are in use (2-5 only)
        if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_2) == LOW ||
            digitalRead(LEGACY_HARDWARE_SELECT_PIN_3) == LOW ||
            digitalRead(LEGACY_HARDWARE_SELECT_PIN_4) == LOW ||
            digitalRead(LEGACY_HARDWARE_SELECT_PIN_5) == LOW)
        {
            // legacy spec
            if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_1) == HIGH)
            {
                if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_2) == LOW)
                    i2cAddress = 8;
                else if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_3) == LOW)
                    i2cAddress = 10;
                else if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_4) == LOW)
                    i2cAddress = 12;
                else if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_5) == LOW)
                    i2cAddress = 14;
            }
            else
            {
                if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_2) == LOW)
                    i2cAddress = 16;
                else if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_3) == LOW)
                    i2cAddress = 18;
                else if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_4) == LOW)
                    i2cAddress = 20;
                else if (digitalRead(LEGACY_HARDWARE_SELECT_PIN_5) == LOW)
                    i2cAddress = 22;
            }
        }
        else
        {   // use standard selection
            i2cAddress = 0;
            if (digitalRead(HARDWARE_SELECT_PIN_1) == LOW)
                i2cAddress |= 1;
            if (digitalRead(HARDWARE_SELECT_PIN_2) == LOW)
                i2cAddress |= 2;
            if (digitalRead(HARDWARE_SELECT_PIN_3) == LOW)
                i2cAddress |= 4;
            i2cAddress = 8 + (i2cAddress * 2);
        }
    }
#endif

public:
    AvrHardware() : Hardware(HIGH,LOW) {
    }
    void init()
    {
        Hardware::init();
        uint16_t clkPin = RX5808_CLK_PIN;    //SEL (CH2) output line to RX5808 module
        uint16_t dataPin = RX5808_DATA_PIN;  //DATA (CH1) output line to RX5808 module
        spi.init(clkPin, dataPin);

        // init pin used to reset paired Arduino via RESET_PAIRED_NODE command
        pinMode(NODE_RESET_PIN, INPUT_PULLUP);

#ifdef ENABLE_SERIAL_PIN
        // init pin that can be pulled low (to GND) to disable serial port
        pinMode(ENABLE_SERIAL_PIN, INPUT_PULLUP);
#endif

    #if defined(USE_I2C) && (!defined(NODE_NUMBER) || !NODE_NUMBER)
        configI2cAddress();
    #else
        delay(100);  // delay a bit a let pin level settle before reading input
    #endif

#ifdef ENABLE_SERIAL_PIN
        if (digitalRead(ENABLE_SERIAL_PIN) == HIGH)
#endif
        {
            initSerial(SERIALCOM, SERIAL_BAUD_RATE);
        }

#ifdef USE_I2C
        i2cInit(i2cAddress);
#endif

        // set ADC prescaler to 16 to speedup ADC readings
        // sampling rate is [ADC clock = 16] / [prescaler = 16] / [conversion clock cycles = 13]
        // ~ 76.8 kHz
        // see https://forum.arduino.cc/t/faster-analog-read/6604
        sbi(ADCSRA, ADPS2);
        cbi(ADCSRA, ADPS1);
        cbi(ADCSRA, ADPS0);
    }

    void initRxModule(uint_fast8_t nIdx, RxModule& rx)
    {
        uint16_t selPin = RX5808_SEL_PIN;    //CLK (CH3) output line to RX5808 module
        uint16_t rssiPin = RSSI_INPUT_PIN;   //RSSI input from RX5808
        rx.init(&spi, selPin, rssiPin);
    }

#ifdef USE_I2C
    void processStatusFlags(const mtime_t ms, const uint8_t statusFlags) {
        static bool i2cMonitorEnabledFlag = false;
        static mtime_t i2cMonitorLastResetTime = 0;

        bool anyNodeActive = false;
        for (int_fast8_t i=rssiRxs.getCount()-1; i>=0; i--) {
            RssiNode& node = rssiRxs.getRssiNode(i);
            if (node.active) {
                anyNodeActive = true;
                break;
            }
        }

        if (anyNodeActive)
        {
            if (i2cMonitorEnabledFlag)
            {
                if ((statusFlags & COMM_ACTIVITY) && (statusFlags & SERIAL_CMD_MSG) == 0)
                {  //I2C communications activity detected; update comms monitor time
                    i2cMonitorLastResetTime = ms;
                }
                else
                {
                    const mtime_t durationWithoutComms = ms - i2cMonitorLastResetTime;
                    if (durationWithoutComms > COMMS_MONITOR_TIME_MS)
                    {  //too long since last communications activity detected
                        LOG_DEBUG("Resetting I2C - no activity period: ", durationWithoutComms, DEC);
                        i2cMonitorEnabledFlag = false;
                        // redo init, which should release I2C pins (SDA & SCL) if "stuck"
                        i2cInit(i2cAddress, true);
                    }
                }
            }
            else if ((statusFlags & POLLING) &&
                    (statusFlags & SERIAL_CMD_MSG) == 0)
            {  //if activated and I2C POLLING cmd received then enable comms monitor
                i2cMonitorEnabledFlag = true;
                i2cMonitorLastResetTime = ms;
            }
        }
        else if (i2cMonitorEnabledFlag)
        {
            i2cMonitorEnabledFlag = false;
        }
    }
#endif

    // Node reset for ISP; resets other node wired to this node's reset pin
    void resetPairedNode(bool pinState)
    {
        if (pinState)
        {
            pinMode(NODE_RESET_PIN, INPUT_PULLUP);
        }
        else
        {
            pinMode(NODE_RESET_PIN, OUTPUT);
            digitalWrite(NODE_RESET_PIN, LOW);
        }
    }

    const char* getProcessorType() {
        return "Arduino";
    }

    uint8_t getAddress()
    {
        return i2cAddress;
    }
};
