#include "config.h"
#include "microclock.h"
#include "rssi.h"
#include "commands.h"

extern void handleStatusMessage(uint8_t msgType, uint8_t data);

constexpr uint_fast16_t RSSI_HISTORY_PAYLOAD_SIZE = 16;
constexpr uint_fast8_t SCAN_HISTORY_PAYLOAD_COUNT = 3;
constexpr uint_fast8_t SCAN_HISTORY_PAYLOAD_SIZE = SCAN_HISTORY_PAYLOAD_COUNT*sizeof(FreqRssi);

uint8_t volatile cmdStatusFlags = NO_STATUS;
uint_fast8_t volatile cmdRssiNodeIndex = 0;

void setFrequency(RssiNode& rssiNode, RxModule& rx, freq_t freq) {
    Settings& settings = rssiNode.getSettings();
    if (freq == POWER_OFF_FREQ) {
        rx.powerDown();
        settings.vtxFreq = POWER_OFF_FREQ;
        rssiNode.active = false;
    } else {
        if (rx.isPoweredDown()) {
            rx.reset();  // reset and power-up
        }
        rx.setFrequency(freq);
        freq_t actualFreq = rx.getFrequency();
        if (actualFreq == canonicalFrequency(freq)) {
            settings.vtxFreq = freq;
            rssiNode.resetState(usclock.millis());
            rssiNode.active = true;
        } else {
            LOG_ERROR("Failed to set vrx frequency, currently is ", actualFreq, DEC);
            settings.vtxFreq = ERR_FREQ;
            rssiNode.active = false;
        }
    }
}

int_fast8_t Message::getPayloadSize() const
{
    int_fast8_t size;
    switch (command)
    {
        case WRITE_ENTER_AT_LEVEL:  // lap pass begins when RSSI is at or above this level
        case WRITE_EXIT_AT_LEVEL:  // lap pass ends when RSSI goes below this level
        case WRITE_MODE:
        case FORCE_END_CROSSING:  // kill current crossing flag regardless of RSSI value
        case RESET_PAIRED_NODE:  // reset paired node for ISP
        case WRITE_CURNODE_INDEX:  // index of current node for this processor
        case JUMP_TO_BOOTLOADER:  // jump to bootloader for flash update
            size = 1;
            break;

        case WRITE_FREQUENCY:
        case SEND_STATUS_MESSAGE:
            size = 2;
            break;

        default:  // invalid command
            LOG_ERROR("Invalid write command: ", command, HEX);
            size = -1;
    }
    return size;
}

// Generic IO write command handler
void Message::handleWriteCommand(const CommSource src)
{
    buffer.flipForRead();
    bool activityFlag = true;
    uint8_t statusFlags = cmdStatusFlags;  // non-volatile copy

    switch (command)
    {
        case WRITE_FREQUENCY:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                const freq_t freq = buffer.read16();
                if ((freq >= MIN_FREQ && freq <= MAX_FREQ) || (freq == POWER_OFF_FREQ))
                {
                    RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                    Settings& settings = rssiNode.getSettings();
                    RxModule& rx = rssiRxs.getRxModule(cmdRssiNodeIndex);
                    setFrequency(rssiNode, rx, freq);
                }
            }
            break;

        case WRITE_MODE:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                const uint8_t mode = buffer.read8();
                setMode((Mode)mode);
            }
            break;

        case WRITE_ENTER_AT_LEVEL:  // lap pass begins when RSSI is at or above this level
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                const rssi_t rssiVal = ioBufferReadRssi(buffer);
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                Settings& settings = rssiNode.getSettings();
                settings.enterAtLevel = rssiVal;
            }
            break;

        case WRITE_EXIT_AT_LEVEL:  // lap pass ends when RSSI goes below this level
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                const rssi_t rssiVal = ioBufferReadRssi(buffer);
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                Settings& settings = rssiNode.getSettings();
                settings.exitAtLevel = rssiVal;
            }
            break;

        case WRITE_CURNODE_INDEX:  // index of current node for this processor
            {
                const uint8_t idx = buffer.read8();
                if (idx < rssiRxs.getCount()) {
                    cmdRssiNodeIndex = idx;
                }
            }
            break;

        case SEND_STATUS_MESSAGE:  // status message sent from server to node
            {
                const uint16_t msg = buffer.read16();  // upper byte is message type, lower byte is data
                handleStatusMessage((uint8_t)(msg >> 8), (uint8_t)(msg & 0x00FF));
            }
            break;

        case FORCE_END_CROSSING:  // kill current crossing flag regardless of RSSI value
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                rssiNode.endCrossing(0);
            }
            break;

        case RESET_PAIRED_NODE:  // reset paired node for ISP
            {
                const uint8_t pinState = buffer.read8();
                hardware.resetPairedNode(pinState);
            }
            break;

        case JUMP_TO_BOOTLOADER:  // jump to bootloader for flash update
            hardware.doJumpToBootloader();
            break;

        default:
            LOG_ERROR("Invalid write command: ", command, HEX);
            activityFlag = false;  // not valid activity
    }

    // indicate communications activity detected
    if (activityFlag)
    {
        statusFlags |= COMM_ACTIVITY;
        if (src == SERIAL_SOURCE) {
            statusFlags |= SERIAL_CMD_MSG;
        }
    }
    cmdStatusFlags = statusFlags; // update volatile
}

template <size_t N,size_t T> void ioBufferWriteExtremum(Buffer<N,T>& buf, const Extremum& e, mtime_t now)
{
    ioBufferWriteRssi(buf, e.rssi);
    buf.write16(toDuration(now - e.firstTime));
    buf.write16(e.duration);
}

void Message::setMode(const Mode mode) const
{
    RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
    Settings& settings = rssiNode.getSettings();
    switch (mode) {
        case TIMER:
            rssiNode.setFilter(&(rssiNode.defaultFilter));
            rssiNode.resetState(usclock.millis());
            settings.mode = mode;
            break;
        case SCANNER:
#ifdef SCAN_HISTORY
            rssiNode.scanHistory.clear();
            rssiNode.setFilter(&(rssiNode.medianFilter));
            {
                RxModule& rx = rssiRxs.getRxModule(cmdRssiNodeIndex);
                setFrequency(rssiNode, rx, MIN_SCAN_FREQ);
                if (rssiNode.active) {
                    settings.mode = mode;
                }
            }
#endif
            break;
        case RAW:
#ifdef RSSI_HISTORY
            rssiNode.rssiHistory.clear();
            rssiNode.setFilter(&(rssiNode.noFilter));
            rssiNode.resetState(usclock.millis());
            settings.mode = mode;
#endif
            break;
    }
    rssiNode.resetState(usclock.millis());
}

// Generic IO read command handler
void Message::handleReadCommand(const CommSource src)
{
    buffer.flipForWrite();
    bool activityFlag = true;
    uint8_t statusFlags = cmdStatusFlags;  // non-volatile copy

    switch (command)
    {
        case READ_ADDRESS:
            buffer.write8(hardware.getAddress());
            break;

        case READ_FREQUENCY:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const Settings& settings = rssiNode.getSettings();
                buffer.write16(settings.vtxFreq);
            } else {
                buffer.write16(0xFFFF);
            }
            break;

        case READ_MODE:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const Settings& settings = rssiNode.getSettings();
                buffer.write8(settings.mode);
            } else {
                buffer.write8(0xFF);
            }
            break;

        case READ_ADC:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RxModule& rx = rssiRxs.getRxModule(cmdRssiNodeIndex);
                uint16_t val = analogRead(rx.getRssiPin());
                buffer.write16(val);
            } else {
                buffer.write16(0xFF);
            }
            break;

        case READ_RSSI:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const State& state = rssiNode.getState();
                const LastPass& lastPass = rssiNode.getLastPass();
                ioBufferWriteRssi(buffer, state.rssi);
                buffer.write8(lastPass.lap);
                buffer.write8(rssiNode.isCrossing());
                statusFlags |= POLLING;
            } else {
                ioBufferWriteRssi(buffer, INVALID_RSSI);
                buffer.write8(0);
                buffer.write8(0xFF);
            }
            break;

        case READ_RSSI_STATS:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const State& state = rssiNode.getState();
                ioBufferWriteRssi(buffer, state.nodeRssiPeak);
                ioBufferWriteRssi(buffer, state.nodeRssiNadir);
            } else {
                ioBufferWriteRssi(buffer, INVALID_RSSI);
                ioBufferWriteRssi(buffer, INVALID_RSSI);
            }
            break;

        case READ_LAP_STATS:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const LastPass& lastPass = rssiNode.getLastPass();
                const mtime_t timeNowVal = usclock.millis();
                buffer.write8(lastPass.lap);
                buffer.write16(toDuration(timeNowVal - lastPass.timestamp));  // ms since lap
                ioBufferWriteRssi(buffer, lastPass.rssiPeak);
                ioBufferWriteRssi(buffer, lastPass.rssiNadir);
            } else {
                buffer.write8(0);
                buffer.write16(0xFFFF);
                ioBufferWriteRssi(buffer, INVALID_RSSI);
                ioBufferWriteRssi(buffer, INVALID_RSSI);
            }
            break;

        case READ_ENTER_STATS:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const LastTrigger& lastEnter = rssiNode.getLastEnter();
                const mtime_t timeNowVal = usclock.millis();
                buffer.write8(lastEnter.lap);
                buffer.write16(toDuration(timeNowVal - lastEnter.timestamp));  // ms since enter
                ioBufferWriteRssi(buffer, lastEnter.rssi);
                buffer.write8(lastEnter.lifetime);
            } else {
                buffer.write8(0);
                buffer.write16(0xFFFF);
                ioBufferWriteRssi(buffer, INVALID_RSSI);
                buffer.write8(0);
            }
            break;

        case READ_EXIT_STATS:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const LastTrigger& lastExit = rssiNode.getLastExit();
                const mtime_t timeNowVal = usclock.millis();
                buffer.write8(lastExit.lap);
                buffer.write16(toDuration(timeNowVal - lastExit.timestamp));  // ms since exit
                ioBufferWriteRssi(buffer, lastExit.rssi);
                buffer.write8(lastExit.lifetime);
            } else {
                buffer.write8(0);
                buffer.write16(0xFFFF);
                ioBufferWriteRssi(buffer, INVALID_RSSI);
                buffer.write8(0);
            }
            break;

        case READ_ANALYTICS:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const State& state = rssiNode.getState();
                History& history = rssiNode.getHistory();
                buffer.write8(state.lifetime);  // NB: signed
                buffer.write16(toDuration(state.loopTimeMicros));
                const ExtremumType extremumType = history.nextToSendType();
                if (extremumType != NONE) {
                    const Extremum extremum = history.popNextToSend();
                    const mtime_t timeNowVal = usclock.millis();
                    ioBufferWriteExtremum(buffer, extremum, timeNowVal);
                } else {
                    ioBufferWriteRssi(buffer, INVALID_RSSI);
                    buffer.write16(0);
                    buffer.write16(0);
                }
            } else {
                buffer.write8(0);
                buffer.write16(0);
                ioBufferWriteRssi(buffer, INVALID_RSSI);
                buffer.write16(0);
                buffer.write16(0);
            }
            break;

        case READ_ENTER_AT_LEVEL:  // lap pass begins when RSSI is at or above this level
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const Settings& settings = rssiNode.getSettings();
                ioBufferWriteRssi(buffer, settings.enterAtLevel);
            } else {
                ioBufferWriteRssi(buffer, INVALID_RSSI);
            }
            break;

        case READ_EXIT_AT_LEVEL:  // lap pass ends when RSSI goes below this level
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                const Settings& settings = rssiNode.getSettings();
                ioBufferWriteRssi(buffer, settings.exitAtLevel);
            } else {
                ioBufferWriteRssi(buffer, INVALID_RSSI);
            }
            break;

        case READ_REVISION_CODE:  // reply with NODE_API_LEVEL and verification value
            buffer.write16((0x25 << 8) + (uint16_t)NODE_API_LEVEL);
            break;

        case READ_RSSI_HISTORY:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                handleReadRssiHistory(rssiNode);
                statusFlags |= POLLING;
            }
            break;

        case READ_SCAN_HISTORY:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                RssiNode& rssiNode = rssiRxs.getRssiNode(cmdRssiNodeIndex);
                handleReadScanHistory(rssiNode);
                statusFlags |= POLLING;
            }
            break;

        case READ_TIME_MILLIS:
            buffer.write32(usclock.millis());
            break;

        case READ_RHFEAT_FLAGS:  // reply with feature flags value
            buffer.write16(hardware.getFeatureFlags());
            break;

        case READ_MULTINODE_COUNT:
            buffer.write8(rssiRxs.getCount());
            break;

        case READ_CURNODE_INDEX:
            if (cmdRssiNodeIndex < rssiRxs.getCount())
            {
                buffer.write8(cmdRssiNodeIndex);
            } else {
                buffer.write8(0xFF);
            }
            break;

        case READ_FW_VERSION:
            buffer.writeText(FIRMWARE_VERSION);
            break;

        case READ_FW_BUILDDATE:
            buffer.writeText(__DATE__);
            break;

        case READ_FW_BUILDTIME:
            buffer.writeText(__TIME__);
            break;

        case READ_FW_PROCTYPE:
            buffer.writeText(hardware.getProcessorType());
            break;

        default:  // If an invalid command is sent, write nothing back, master must react
            LOG_ERROR("Invalid read command: ", command, HEX);
            activityFlag = false;  // not valid activity
    }

    // indicate communications activity detected
    if (activityFlag)
    {
        statusFlags |= COMM_ACTIVITY;
        if (src == SERIAL_SOURCE) {
            statusFlags |= SERIAL_CMD_MSG;
        }
    }
    cmdStatusFlags = statusFlags;  // update volatile

    if (!buffer.isEmpty())
    {
        buffer.writeChecksum();
    }
}

void Message::handleReadRssiHistory(RssiNode& rssiNode)
{
    int i = 0;
#ifdef RSSI_HISTORY
    CircularBuffer<volatile rssi_t,RSSI_HISTORY_SIZE>& rssiHistory = rssiNode.rssiHistory;
    const uint_fast16_t n = min(rssiHistory.size(), RSSI_HISTORY_PAYLOAD_SIZE);
    for (; i<n; i++) {
        ioBufferWriteRssi(buffer, rssiHistory.shift());
    }
    if (i<RSSI_HISTORY_PAYLOAD_SIZE && rssiNode.rssiHistoryComplete) {
        ioBufferWriteRssi(buffer, INVALID_RSSI);
        i++;
        rssiNode.rssiHistoryComplete = false;
    }
#endif
    for (; i<RSSI_HISTORY_PAYLOAD_SIZE; i++) {
        ioBufferWriteRssi(buffer, INVALID_RSSI);
    }
}

void Message::handleReadScanHistory(RssiNode& rssiNode)
{
    int i = 0;
#ifdef SCAN_HISTORY
    CircularBuffer<FreqRssi,SCAN_HISTORY_SIZE>& scanHistory = rssiNode.scanHistory;
    const uint_fast8_t n = min(scanHistory.size(), SCAN_HISTORY_PAYLOAD_COUNT);
    for (; i<n; i++) {
        ioBufferWriteFreqRssi(buffer, scanHistory.shift());
    }
#endif
    const FreqRssi f_r = {0, INVALID_RSSI};
    for (; i<SCAN_HISTORY_PAYLOAD_COUNT; i++) {
        ioBufferWriteFreqRssi(buffer, f_r);
    }
}


void handleStreamEvent(Stream& stream, Message& msg, const CommSource src)
{
    const uint8_t nextByte = stream.read();
    if (msg.buffer.size == 0)
    {
        // new command
        msg.command = (Command) nextByte;
        if (msg.isWriteCommand())
        {
            const int_fast8_t payloadSize = msg.getPayloadSize();
            if (payloadSize > 0)
            {
                msg.buffer.index = 0;
                msg.buffer.size = payloadSize + 1;  // include checksum byte
            }
        }
        else
        {
            msg.handleReadCommand(src);
            sendReadCommandResponse(stream, msg, src);
        }
    }
    else
    {
        // existing command
        msg.buffer.data[msg.buffer.index++] = nextByte;
        validateAndProcessWriteCommand(msg, src);
    }
}

void sendReadCommandResponse(Stream& stream, Message& msg, const CommSource src) {
    // if there is pending data, send it
    if (msg.buffer.size > 0) {
// don't write to serial in debug mode
#ifdef DEBUG
        if (src != SERIAL_SOURCE)
#endif
        {
        stream.write(msg.buffer.data, msg.buffer.size);
        }
        msg.buffer.size = 0;
    }
    msg.command = INVALID_COMMAND;  // Clear previous command
}

void validateAndProcessWriteCommand(Message& msg, const CommSource src) {
    if (msg.buffer.index == msg.buffer.size)
    {
        const uint8_t checksum = msg.buffer.calculateChecksum(msg.buffer.size - 1);
        if (msg.buffer.data[msg.buffer.size - 1] == checksum)
        {
            msg.handleWriteCommand(src);
        }
        else
        {
            LOG_ERROR("Invalid checksum", checksum, HEX);
        }
        msg.buffer.size = 0;
        msg.command = INVALID_COMMAND;  // Clear previous command
    }
}
