#include "config.h"
#include "rssi.h"
#include "util/persistent-homology.h"
#ifdef USE_MQTT
#include "mqtt.h"
#include <stdio.h>
#include <sys/time.h>
#include "json.h"
#endif

constexpr uint16_t MIN_TUNETIME = 35;  // after set freq need to wait this long before reading RSSI

inline void initExtremum(Extremum& e, const rssi_t rssi, const mtime_t ts)
{
    e.rssi = rssi;
    e.firstTime = ts;
    e.duration = 0;
}

RssiNode::RssiNode()
{
    setFilter(&defaultFilter);
}

void RssiNode::setFilter(Filter<rssi_t> *f)
{
    filter = f;
}

void RssiNode::start(const mtime_t ms, const utime_t us)
{
    lastResetTimeMs = ms;
    state.lastloopMicros = us;
}

bool RssiNode::isStateValid() const
{
    return state.nodeRssiNadir <= state.rssi && state.rssi <= state.nodeRssiPeak;
}

void RssiNode::resetState(const mtime_t ms)
{
    filter->reset();
    state.reset();
    history.reset();
    needsToSettle = true;
    lastResetTimeMs = ms;
}

RssiResult RssiNode::process(const rssi_t rssi, const mtime_t ms)
{
    RssiResult result = {0};

    currentMillis = ms;

    filter->addRawValue(ms, rssi);
    if (filter->isFilled()) {
        const int rssiChange = state.readRssiFromFilter(filter);

        LOG_DEBUG("RSSI timestamp: ", state.rssiTimestamp, DEC);
        LOG_DEBUG("RSSI: ", state.rssi, DEC);
        LOG_DEBUG("RSSI change: ", rssiChange, DEC);

        // wait until after-tune-delay time is fulfilled
        if (!needsToSettle)
        {  //don't start operations until after first WRITE_FREQUENCY command is received

            state.updateRssiStats();
            switch (settings.mode) {
                case TIMER:
                    result.crossing = timerHandler(rssiChange);
                    break;
                case SCANNER:
                    result.nextFreq = scannerHandler(rssiChange);
                    break;
                case RAW:
                    rawHandler(rssiChange);
                    break;
            }
        }

        // check if RSSI is stable after tune
        if (needsToSettle && (ms - lastResetTimeMs) >= MIN_TUNETIME)
        {
            needsToSettle = false;  // don't need to check again until next freq change
        }
    }

    return result;
}

bool RssiNode::timerHandler(const int rssiChange) {
    const ExtremumType currentType = updateHistory(rssiChange);
    const bool crossing = checkForCrossing(currentType, rssiChange);

    /*** pass processing **/

    if (crossing)
    {  //lap pass is in progress
        // Find the peak rssi and the time it occured during a crossing event
        if (state.rssi > state.passPeak.rssi)
        {
            // this is first time this peak RSSI value was seen, so save value and timestamp
            initExtremum(state.passPeak, state.rssi, state.rssiTimestamp);
        }
        else if (state.rssi == state.passPeak.rssi)
        {
            // if at max peak for more than one iteration then track duration
            // so middle-timestamp value can be returned
            state.passPeak.duration = toDuration(state.rssiTimestamp - state.passPeak.firstTime);
        }
    }
    else
    {
        // track lowest RSSI seen between passes
        state.passRssiNadir = min((rssi_t)state.rssi, (rssi_t)state.passRssiNadir);
    }

    return crossing;
}

ExtremumType RssiNode::updateHistory(const int rssiChange)
{
    ExtremumType etype;
    if (rssiChange > 0)
    {  // RSSI is rising
        // whenever history is rising, record the time and value as a peak
        etype = history.startNewPeak(state.rssi, state.rssiTimestamp);

        // if RSSI was falling or unchanged, but it's rising now, we found a nadir
        // copy the values to be sent in the next loop
        history.checkForNadir(rssiChange);
    }
    else if (rssiChange < 0)
    {  // RSSI is falling
        // whenever history is falling, record the time and value as a nadir
        etype = history.startNewNadir(state.rssi, state.rssiTimestamp);

        // if RSSI was rising or unchanged, but it's falling now, we found a peak
        // copy the values to be sent in the next loop
        history.checkForPeak(rssiChange);
    }
    else
    {  // RSSI is equal
        switch (history.extremumType) {
            case PEAK:
                history.peak.duration = toDuration(state.rssiTimestamp - history.peak.firstTime);
                if (history.peak.duration == MAX_DURATION)
                {
                    history.startNewPeak(state.rssi, state.rssiTimestamp);
                }
                break;
            case NADIR:
                history.nadir.duration = toDuration(state.rssiTimestamp - history.nadir.firstTime);
                if (history.nadir.duration == MAX_DURATION)
                {
                    history.startNewNadir(state.rssi, state.rssiTimestamp);
                }
                break;
            default:
                initExtremum(history.peak, state.rssi, state.rssiTimestamp);
                initExtremum(history.nadir, state.rssi, state.rssiTimestamp);
                break;
        }
        etype = history.extremumType;
    }

    history.recordRssiChange(rssiChange);

    // try to buffer latest peak/nadir (don't overwrite any unsent peak/nadir)
    history.bufferPeak();
    history.bufferNadir();

    return etype;
}

bool RssiNode::checkForCrossing(const ExtremumType t, const int rssiChange)
{
#if defined(USE_PH)
    const bool isCrossing = checkForCrossing_ph(t, settings.enterAtLevel, settings.exitAtLevel);
#elif defined(__TEST__)
    bool isCrossing;
    if (settings.usePh) {
        isCrossing = checkForCrossing_ph(t, settings.enterAtLevel, settings.exitAtLevel);
    } else {
        isCrossing = checkForCrossing_old(settings.enterAtLevel, settings.exitAtLevel);
    }
#else
    const bool isCrossing = checkForCrossing_old(settings.enterAtLevel, settings.exitAtLevel);
#endif
    LOG_DEBUG("Is crossing: ", isCrossing, DEC);
    return isCrossing;
}

#if defined(USE_PH) || defined(__TEST__)
bool RssiNode::checkForCrossing_ph(const ExtremumType currentType, const uint8_t enterThreshold, const uint8_t exitThreshold)
{
    const SENDBUFFER& sendBuffer = *((SENDBUFFER*)(history.sendBuffer));

    if (currentType == NONE || sendBuffer.size() == 0) {
        return state.crossing;
    }

#if defined(USE_MQTT) || defined(DEBUG)
    bool triggered = false;
#endif
    const ExtremumType prevType = sendBuffer.typeAt(sendBuffer.size()-1);
    if (state.crossing && prevType == PEAK && currentType == NADIR) {
        int_fast8_t lastIdx = preparePhData(history.nadir.rssi);
        calculateNadirPersistentHomology<rssi_t,PH_HISTORY_SIZE>(phData, phSortedIdxs, lastIdx+1, ccs, &lastIdx);

        // find lifetime of last value when a nadir
        if (lastIdx < 0) {
            const ConnectedComponent& cc = ccs[-lastIdx-1];
            const uint_fast8_t lastLifetime = cc.nadirLifetime(phData);
            state.lifetime = -constrain((int)lastLifetime, 0, 127);
            if (lastLifetime > exitThreshold) {
                endCrossing(lastLifetime);
#if defined(USE_MQTT) || defined(DEBUG)
                triggered = true;
#endif
            }
        }
    } else if (!state.crossing && prevType == NADIR && currentType == PEAK) {
        int_fast8_t lastIdx = preparePhData(history.peak.rssi);
        calculatePeakPersistentHomology<rssi_t,PH_HISTORY_SIZE>(phData, phSortedIdxs, lastIdx+1, ccs, &lastIdx);

        // find lifetime of last value when a peak
        if (lastIdx < 0) {
            const ConnectedComponent& cc = ccs[-lastIdx-1];
            const uint_fast8_t lastLifetime = cc.peakLifetime(phData);
            state.lifetime = constrain((int)lastLifetime, 0, 127);
            if (lastLifetime > enterThreshold) {
                startCrossing(lastLifetime);
#if defined(USE_MQTT) || defined(DEBUG)
                triggered = true;
#endif
            }
        }
    }

    LOG_DEBUG("Lifetime: ", state.lifetime, DEC);
    LOG_DEBUG("Is triggered: ", triggered, DEC);

#ifdef USE_MQTT
    if (!triggered && state.lifetime != 0 && (state.rssiTimestamp & MQTT_SAMPLE_INTERVAL_MASK) == 0)
    {
        char json[64] = "";
        int n = jsonStart(json);
        n += jsonPropertyUInt(json, "rssi", state.rssi);
        n += jsonPropertyUInt(json, "timestamp", state.rssiTimestamp);
        n += jsonPropertyUInt(json, "lifetime", state.lifetime);
#ifdef USE_NTP
        timeval currentTime;
        gettimeofday(&currentTime, NULL);
        n += jsonPropertyTime(json, "utc", &currentTime);
#endif
        n += jsonEnd(json);
        mqttPublish(*this, "sample", json, n);
    }
#endif

    return state.crossing;
}

int_fast8_t RssiNode::preparePhData(const rssi_t currentValue)
{
    const SENDBUFFER& sendBuffer = *((SENDBUFFER*)(history.sendBuffer));

    // copy history to fast array
    sendBuffer.copyRssi(phData);
    const int_fast8_t lastIdx = sendBuffer.size();
    // insert current value
    phData[lastIdx] = currentValue;

    // sort phData
    int_fast8_t j = lastIdx-1;
    for (; j>=0 && phData[sendBuffer.sortedIdxs[j]] > currentValue; j--) {
        phSortedIdxs[j+1] = sendBuffer.sortedIdxs[j];
    }
    phSortedIdxs[j+1] = lastIdx;
    for (; j>=0; j--) {
        phSortedIdxs[j] = sendBuffer.sortedIdxs[j];
    }

    return lastIdx;
}
#endif

bool RssiNode::checkForCrossing_old(const rssi_t enterThreshold, const rssi_t exitThreshold)
{
    /*** crossing transition ***/

    if ((!state.crossing) && state.rssi >= enterThreshold)
    {
        // quad is going through the gate (lap pass starting)
        startCrossing(state.rssi);
    }
    else if (state.crossing && state.rssi < exitThreshold)
    {
        // quad has left the gate
        endCrossing(state.rssi);
    }
#ifdef USE_MQTT
    else if ((state.rssiTimestamp & MQTT_SAMPLE_INTERVAL_MASK) == 0)
    {
        char json[64] = "";
        int n = jsonStart(json);
        n += jsonPropertyUInt(json, "rssi", state.rssi);
        n += jsonPropertyUInt(json, "timestamp", state.rssiTimestamp);
#ifdef USE_NTP
        timeval currentTime;
        gettimeofday(&currentTime, NULL);
        n += jsonPropertyTime(json, "utc", &currentTime);
        n += jsonPropertyULong(json, "clock", currentMillis);
#endif
        n += jsonEnd(json);
        mqttPublish(*this, "sample", json, n);
    }
#endif

    return state.crossing;
}

bool RssiNode::isCrossing() const
{
    return state.crossing;
}

void RssiNode::startCrossing(const uint8_t trigger)
{
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        lastEnter.rssi = state.rssi;
        lastEnter.timestamp = state.rssiTimestamp;
#ifdef USE_PH
        lastEnter.lifetime = trigger;
#endif
        lastEnter.lap = lastPass.lap + 1;
    }

    state.crossing = true;

#ifdef USE_MQTT
    publishTrigger("enter", lastEnter);
#endif
}

/*** Function called when crossing ends (by RSSI or I2C command). */
void RssiNode::endCrossing(const uint8_t trigger)
{
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        lastExit.rssi = state.rssi;
        lastExit.timestamp = state.rssiTimestamp;
#ifdef USE_PH
        lastExit.lifetime = trigger;
#endif
        lastExit.lap = lastPass.lap + 1;
    }

    // save values for lap pass
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        lastPass.rssiPeak = state.passPeak.rssi;
        // lap timestamp is between first and last peak RSSI
        lastPass.timestamp = state.passPeak.firstTime + state.passPeak.duration / 2;
        lastPass.rssiNadir = state.passRssiNadir;
        lastPass.lap = lastPass.lap + 1;
    }

    // reset lap-pass variables
    state.resetPass();

#ifdef USE_MQTT
    publishTrigger("exit", lastExit);

    char json[64] = "";
    int n = jsonStart(json);
    n += jsonPropertyUInt(json, "lap", lastPass.lap);
    n += jsonPropertyUInt(json, "rssi", lastPass.rssiPeak);
    n += jsonPropertyUInt(json, "timestamp", lastPass.timestamp);
#ifdef USE_NTP
    timeval currentTime;
    gettimeofday(&currentTime, NULL);
    n += jsonPropertyTime(json, "utc", &currentTime);
    n += jsonPropertyULong(json, "clock", currentMillis);
#endif
    n += jsonEnd(json);
    mqttPublish(*this, "pass", json, n);
#endif
}

#ifdef USE_MQTT
void RssiNode::publishTrigger(const char*const topic, const LastTrigger& trigger)
{
    char json[64] = "";
    int n = jsonStart(json);
    n += jsonPropertyUInt(json, "lap", trigger.lap);
    n += jsonPropertyUInt(json, "rssi", trigger.rssi);
    n += jsonPropertyUInt(json, "timestamp", trigger.timestamp);
    #if defined(USE_PH)
    n += jsonPropertyUInt(json, "lifetime", trigger.lifetime);
    #endif
    #ifdef USE_NTP
    timeval currentTime;
    gettimeofday(&currentTime, NULL);
    n += jsonPropertyTime(json, "utc", &currentTime);
    n += jsonPropertyULong(json, "clock", currentMillis);
    #endif
    n += jsonEnd(json);
    mqttPublish(*this, topic, json, n);
}
#endif

freq_t RssiNode::scannerHandler(const int rssiChange)
{
#ifdef SCAN_HISTORY
    if (!scanHistory.isFull())
    {
        const freq_t freq = settings.vtxFreq;
        const FreqRssi f_r = {freq, state.rssi};
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
        {
            scanHistory.push(f_r);
        }

        freq_t nextFreq = freq + SCAN_FREQ_INCR;
        if (nextFreq > MAX_SCAN_FREQ) {
            nextFreq = MIN_SCAN_FREQ;
        }
        return nextFreq;
    } else {
        return 0;
    }
#endif
}

void RssiNode::rawHandler(const int rssiChange)
{
#ifdef RSSI_HISTORY
    if (!rssiHistoryComplete) {
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
        {
            if (!rssiHistory.isFull()) {
                rssiHistory.push(state.rssi);
            } else {
                rssiHistoryComplete = true;
            }
        }
    }
#endif
}


int State::readRssiFromFilter(Filter<rssi_t>* filter)
{
    lastRssi = rssi;
    rssi = filter->getFilteredValue();
    rssiTimestamp = filter->getFilterTimestamp();

    // ensure signed arithmetic
    const int delta = (int)rssi - (int)lastRssi;
    return delta;
}

/*** node lifetime RSSI max/min ***/
void State::updateRssiStats() {
    if (rssi > nodeRssiPeak)
    {
        nodeRssiPeak = rssi;
    }

    if (rssi < nodeRssiNadir)
    {
        nodeRssiNadir = rssi;
    }
}

/*** Calculate the time it takes to run the main loop. */
void State::updateLoopTime(const utime_t loopMicros)
{
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        loopTimeMicros = loopMicros - lastloopMicros;
    }
    lastloopMicros = loopMicros;
}

void State::resetPass()
{
    crossing = false;
    invalidatePeak(passPeak);
    passRssiNadir = MAX_RSSI;
}

void State::reset()
{
    rssi = INVALID_RSSI;
    resetPass();
    nodeRssiPeak = 0;
    nodeRssiNadir = MAX_RSSI;
}

ExtremumType History::startNewPeak(const rssi_t rssi, const mtime_t ts) {
  // must buffer latest peak to prevent losing it
  bufferPeak(true);
  // reset peak
  initExtremum(peak, rssi, ts);
  extremumType = PEAK;
  return extremumType;
}

ExtremumType History::startNewNadir(const rssi_t rssi, const mtime_t ts) {
  // must buffer latest nadir to prevent losing it
  bufferNadir(true);
  // reset nadir
  initExtremum(nadir, rssi, ts);
  extremumType = NADIR;
  return extremumType;
}

void History::bufferPeak(const bool force)
{
    if (hasPending & PENDING_PEAK)
    {
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
        {
            const bool buffered = sendBuffer->addPeak(peak, force);
            if (buffered || force)
            {
                hasPending &= ~PENDING_PEAK;
            }
        }
    }
}

void History::bufferNadir(const bool force)
{
    if (hasPending & PENDING_NADIR)
    {
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
        {
            const bool buffered = sendBuffer->addNadir(nadir, force);
            if (buffered || force)
            {
                hasPending &= ~PENDING_NADIR;
            }
        }
    }
}

void History::recordRssiChange(const int delta)
{
  // clamp to prevent overflow
  prevRssiChange = constrain(delta, -127, 127);
}

ExtremumType History::nextToSendType() const
{
    ExtremumType t;
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        t = sendBuffer->nextType();
    }
    return t;
}

Extremum History::popNextToSend()
{
    Extremum e;
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        e = sendBuffer->popNext();
    }
    return e;
}

bool History::checkForPeak(const int rssiChange)
{
  if (prevRssiChange >= 0 && rssiChange < 0 && isPeakValid(peak))
  {  // was rising or unchanged
      // declare a new peak
      hasPending |= PENDING_PEAK;
      return true;
  }
  else
  {
      return false;
  }
}

bool History::checkForNadir(const int rssiChange)
{
  if (prevRssiChange <= 0 && rssiChange > 0 && isNadirValid(nadir))
  {  // was falling or unchanged
      // declare a new nadir
      hasPending |= PENDING_NADIR;
      return true;
  }
  else
  {
      return false;
  }
}

void History::reset()
{
    extremumType = NONE;
    invalidatePeak(peak);
    invalidateNadir(nadir);
    hasPending = PENDING_NONE;
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        sendBuffer->clear();
    }
    prevRssiChange = 0;
}
