import gevent
import bisect
from collections import namedtuple
from typing import List, Tuple


RssiSample = namedtuple('RssiSample', ['timestamp', 'rssi'], defaults=[0,0])
LifetimeSample = namedtuple('LifetimeSample', ['timestamp', 'lifetime'], defaults=[0,0])


def unpack_8(data: bytes) -> int:
    return data[0]


def unpack_8_signed(data: bytes) -> int:
    v = data[0]
    mag = v & 0x7F
    if v & 0x80:
        return mag - 128
    else:
        return mag


def pack_8(data: int):
    return [int(data & 0xFF)]


def unpack_16(data: bytes) -> int:
    '''Returns the full variable from 2 bytes input.'''
    result = data[0]
    result = (result << 8) | data[1]
    return result


def pack_16(data: int):
    '''Returns a 2 part array from the full variable.'''
    part_a = (data >> 8) & 0xFF
    part_b = (data & 0xFF)
    return [int(part_a), int(part_b)]


def unpack_32(data: bytes) -> int:
    '''Returns the full variable from 4 bytes input.'''
    result = data[0]
    result = (result << 8) | data[1]
    result = (result << 8) | data[2]
    result = (result << 8) | data[3]
    return result


def pack_32(data: int):
    '''Returns a 4 part array from the full variable.'''
    part_a = (data >> 24) & 0xFF
    part_b = (data >> 16) & 0xFF
    part_c = (data >> 8) & 0xFF
    part_d = (data & 0xFF)
    return [int(part_a), int(part_b), int(part_c), int(part_d)]


def calculate_checksum(data: bytearray):
    checksum = sum(data) & 0xFF
    return checksum


def ensure_iter(l):
    if type(l) == str or not hasattr(l, '__iter__'):
        l = [l]
    return l


class SampleHistory:
    def __init__(self):
        # timestamps (ms)
        self._times: List[int] = []
        self._values: List[int] = []
        self.lock = gevent.lock.RLock()

    def __len__(self):
        assert len(self._times) == len(self._values)
        return len(self._times)

    def append(self, ts: int, value: int):
        with self.lock:
            n = len(self._times)
            # if previous two entries have same value then just extend time on last entry
            if n >= 2 and self._values[n-1] == value and self._values[n-2] == value:
                self._times[n-1] = ts
            else:
                self._times.append(ts)
                self._values.append(value)

    def merge(self, new_entries: List[Tuple[int, int]]):
        with self.lock:
            for ts_value in new_entries:
                idx = bisect.bisect_left(self._times, ts_value[0])
                if idx < len(self._times) and ts_value[0] == self._times[idx]:
                    # replace existing value
                    self._values[idx] = ts_value[1]
                else:
                    self._times.insert(idx, ts_value[0])
                    self._values.insert(idx, ts_value[1])
            
    def set(self, times, values):
        if len(times) != len(values):
            raise ValueError("History time and value lists must have the same length")
        with self.lock:
            self._times = times
            self._values = values

    def get(self):
        with self.lock:
            return self._times.copy(), self._values.copy()

    def prune(self, keep_after):
        with self.lock:
            prune_idx = bisect.bisect_left(self._times, keep_after)
            del self._values[:prune_idx]
            del self._times[:prune_idx]


class ExtremumFilter:
    def __init__(self):
        self.prev_t = None
        self.prev_x = None
        self.delta = None

    def filter(self, t: int, x: int):
        '''Includes inflexion points'''
        if self.prev_x is not None:
            new_delta = x - self.prev_x
            if self.delta > 0 and new_delta <= 0:
                next_x = self.prev_x
            elif self.delta < 0 and new_delta >= 0:
                next_x = self.prev_x
            elif self.delta == 0 and new_delta != 0:
                next_x = self.prev_x
            else:
                next_x = None
        else:
            new_delta = 0
            next_x = self.prev_x
        next_t = self.prev_t
        self.prev_t = t
        self.prev_x = x
        self.delta = new_delta
        return next_t, next_x

