import unittest
from rh.interface.ChorusInterface import ChorusInterface
from rh.interface.MockInterface import MockInterface
from rh.interface import RssiSample
from rh.apis.chorus_api import ChorusAPI
from . import stub_sensor
import gevent

class ChorusTest(unittest.TestCase):
    class DummySerial:
        def __init__(self, handler):
            self.port = 'COM'
            self.handler = handler
            self.buffer = []

        def write(self, raw_data):
            msgs = bytes.decode(raw_data)[:-1]
            for msg in msgs.split('\n'):
                response = self.handler(msg)
                self.buffer.append(response)

        def read_until(self):
            while not self.buffer:
                gevent.sleep(0.1)
            data = self.buffer.pop(0)
            return data

    def test(self):
        mock_intf = MockInterface()
        started = False
        race_stopped = False
        laps = 0
        def on_start():
            nonlocal started
            started = True
        def on_stop_race():
            nonlocal race_stopped
            race_stopped = True
        def on_reset_race():
            pass
        def on_pass(node, lap_ts, source, rssi):
            nonlocal laps
            laps += 1

        api = ChorusAPI(None, mock_intf, stub_sensor.discover(), on_start, on_stop_race, on_reset_race)
        api_io = ChorusTest.DummySerial(lambda data : api._process_message(data))
        intf = ChorusInterface(api_io)
        api.serial_io = ChorusTest.DummySerial(lambda data : intf._process_message(intf.node_managers[0], data))
        intf.listener.on_pass = on_pass
        self.assertTrue(started)
        for sensor in intf.sensors:
            sensor.update()
            intf._update()
            self.assertGreater(len(sensor.getReadings()), 0)
        intf.set_frequency(2, 5885)
        self.assertEqual(mock_intf.nodes[2].frequency, 5885)
        intf.set_enter_at_level(4, 33)
        self.assertEqual(mock_intf.nodes[4].enter_at_level, 33)
        self.assertEqual(mock_intf.nodes[4].exit_at_level, 33)
        intf.set_exit_at_level(5, 34)
        self.assertEqual(mock_intf.nodes[4].enter_at_level, 33)
        self.assertEqual(mock_intf.nodes[4].exit_at_level, 33)
        self.assertEqual(mock_intf.nodes[5].enter_at_level, 34)
        self.assertEqual(mock_intf.nodes[5].exit_at_level, 34)
        api.emit_pass_record(mock_intf.nodes[0], 1, 98)
        gevent.sleep(0)
        self.assertEqual(laps, 1)

        intf.start()
        sample_0 = RssiSample(0, 66)  # timestamp value is unused
        sample_3 = RssiSample(0, 67)  # timestamp value is unused
        mock_intf.nodes[0].current_rssi = sample_0
        mock_intf.nodes[3].current_rssi = sample_3
        api.emit_rssi('*')
        gevent.sleep(0.5)
        intf.stop()
        self.assertEqual(intf.nodes[0].current_rssi.rssi, sample_0.rssi)
        self.assertEqual(intf.nodes[3].current_rssi.rssi, sample_3.rssi)


if __name__ == '__main__':
    unittest.main()
